@@ -41,17 +41,18 @@ def basename(filename):
     #we can't use os.path.basename,
     #because the remote end may have sent us a filename
     #which is using a different pathsep
+    tmp = filename
     for sep in ("\\", "/", os.sep):
-        i = filename.rfind(sep) + 1
-        filename = filename[i:]
-    if not WIN32:
-        return filename
-    #many characters aren't allowed at all on win32:
-    tmp = ""
-    for char in filename:
-        if ord(char)<32 or char in ("<", ">", ":", "\"", "|", "?", "*"):
-            char = "_"
-        tmp += char
+        i = tmp.rfind(sep) + 1
+        tmp = tmp[i:]
+    filename = tmp
+    if WIN32:   # pragma: no cover
+        #many characters aren't allowed at all on win32:
+        tmp = ""
+        for char in filename:
+            if ord(char)<32 or char in ("<", ">", ":", "\"", "|", "?", "*"):
+                char = "_"
+            tmp += char
     return tmp
 
 def safe_open_download_file(basefilename, mimetype):
@@ -83,7 +84,7 @@ def safe_open_download_file(basefilename, mimetype):
     return filename, fd
 
 
-class FileTransferAttributes(object):
+class FileTransferAttributes:
 
     def __init__(self):
         self.init_attributes()
@@ -120,7 +121,7 @@ class FileTransferAttributes(object):
         self.open_command = open_command
         filelog("file transfer attributes=%s", self.get_file_transfer_features())
 
-    def get_file_transfer_features(self):
+    def get_file_transfer_features(self) -> dict:
         #used in hello packets
         return {
                 "file-transfer"     : self.file_transfer,
@@ -136,7 +137,7 @@ class FileTransferAttributes(object):
                 "file-ask-timeout"  : self.file_ask_timeout,
                 }
 
-    def get_info(self):
+    def get_info(self) -> dict:
         #slightly different from above... for legacy reasons
         #this one is used for get_info() in a proper "file." namespace from server_base.py
         return {
@@ -161,7 +162,7 @@ class FileTransferHandler(FileTransferAttributes):
     """
 
     def init_attributes(self, *args):
-        FileTransferAttributes.init_attributes(self, *args)
+        super().init_attributes(*args)
         self.remote_file_transfer = False
         self.remote_file_transfer_ask = False
         self.remote_printing = False
@@ -179,11 +180,10 @@ class FileTransferHandler(FileTransferAttributes):
         self.receive_chunks_in_progress = {}
         self.file_descriptors = set()
         if not getattr(self, "timeout_add", None):
-            from xpra.gtk_common.gobject_compat import import_glib
-            glib = import_glib()
-            self.timeout_add = glib.timeout_add
-            self.idle_add = glib.idle_add
-            self.source_remove = glib.source_remove
+            from gi.repository import GLib
+            self.timeout_add = GLib.timeout_add
+            self.idle_add = GLib.idle_add
+            self.source_remove = GLib.source_remove
 
     def cleanup(self):
         for t in self.pending_send_data_timers.values():
@@ -196,7 +196,7 @@ class FileTransferHandler(FileTransferAttributes):
         for x in tuple(self.file_descriptors):
             try:
                 os.close(x)
-            except (OSError, IOError):
+            except OSError:
                 pass
         self.file_descriptors = set()
         self.init_attributes()
@@ -225,7 +225,7 @@ class FileTransferHandler(FileTransferAttributes):
         filelog("file transfer remote caps: open-url=%s        (ask=%s)",
                 self.remote_open_url, self.remote_open_url_ask)
 
-    def get_info(self):
+    def get_info(self) -> dict:
         info = FileTransferAttributes.get_info(self)
         info["remote"] = {
             "file-transfer"     : self.remote_file_transfer,
@@ -262,11 +262,10 @@ class FileTransferHandler(FileTransferAttributes):
 
     def _process_send_file_chunk(self, packet):
         chunk_id, chunk, file_data, has_more = packet[1:5]
-        chunk_id = bytestostr(chunk_id)
         filelog("_process_send_file_chunk%s", (chunk_id, chunk, "%i bytes" % len(file_data), has_more))
         chunk_state = self.receive_chunks_in_progress.get(chunk_id)
         if not chunk_state:
-            filelog.error("Error: cannot find the file transfer id '%s'", nonl(chunk_id))
+            filelog.error("Error: cannot find the file transfer id '%s'", nonl(bytestostr(chunk_id)))
             self.send("ack-file-chunk", chunk_id, False, "file transfer id not found", chunk)
             return
         fd = chunk_state[1]
@@ -292,7 +291,7 @@ class FileTransferHandler(FileTransferAttributes):
             del self.receive_chunks_in_progress[chunk_id]
             try:
                 os.close(fd)
-            except (OSError, IOError):
+            except OSError:
                 pass
             return
         self.send("ack-file-chunk", chunk_id, True, "", chunk)
@@ -318,7 +317,9 @@ class FileTransferHandler(FileTransferAttributes):
         elapsed = monotonic_time()-start_time
         mimetype = bytestostr(mimetype)
         filelog("%i bytes received in %i chunks, took %ims", filesize, chunk, elapsed*1000)
-        self.process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
+        t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False,
+                         args=(filename, mimetype, printit, openit, filesize, options))
+        filelog("started process-download thread: %s", t)
 
     def accept_data(self, send_id, dtype, basefilename, printit, openit):
         #subclasses should check the flags,
@@ -330,24 +331,28 @@ class FileTransferHandler(FileTransferAttributes):
                 self.file_transfer, self.file_transfer_ask)
         filelog("accept_data: open-files=%s, open-files-ask=%s",
                 self.open_files, self.open_files_ask)
-        if printit:
-            return self.printing and not self.printing_ask
         if not self.file_transfer or self.file_transfer_ask:
-            return False
-        if openit:
-            return self.open_files and not self.open_files_ask
-        return True
+            return None
+        if printit and (not self.printing or self.printing_ask):
+            printit = False
+        if openit and (not self.open_files or self.open_files_ask):
+            #we can't ask in this implementation,
+            #so deny the request to open it:
+            openit = False
+        return (printit, openit)
 
     def _process_send_file(self, packet):
         #the remote end is sending us a file
-        start = monotonic_time()
         basefilename, mimetype, printit, openit, filesize, file_data, options = packet[1:8]
         send_id = ""
         if len(packet)>=9:
             send_id = packet[8]
-        if not self.accept_data(send_id, b"file", basefilename, printit, openit):
+        r = self.accept_data(send_id, b"file", basefilename, printit, openit)
+        if r is None:
             filelog.warn("Warning: file transfer rejected for file '%s'", bytestostr(basefilename))
             return
+        #accept_data can override the flags:
+        printit, openit = r
         options = typedict(options)
         if printit:
             l = printlog
@@ -366,7 +371,7 @@ class FileTransferHandler(FileTransferAttributes):
         #basefilename should be utf8:
         try:
             base = basefilename.decode("utf8")
-        except:
+        except UnicodeDecodeError:
             base = bytestostr(basefilename)
         try:
             filename, fd = safe_open_download_file(base, mimetype)
@@ -379,6 +384,7 @@ class FileTransferHandler(FileTransferAttributes):
             return
         self.file_descriptors.add(fd)
         if chunk_id:
+            chunk_id = strtobytes(chunk_id)
             l = len(self.receive_chunks_in_progress)
             if l>=MAX_CONCURRENT_FILES:
                 self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress: %i" % l, 0)
@@ -416,14 +422,7 @@ class FileTransferHandler(FileTransferAttributes):
             os.write(fd, file_data)
         finally:
             os.close(fd)
-        self.transfer_progress_update(False, send_id, monotonic_time()-start, filesize, filesize, None)
-        self.process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
-
-
-    def process_downloaded_file(self, filename, mimetype, printit, openit, filesize, options):
-        t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False,
-                         args=(filename, mimetype, printit, openit, filesize, options))
-        filelog("started process-download thread: %s", t)
+        self.do_process_downloaded_file(filename, mimetype, printit, openit, filesize, options)
 
     def do_process_downloaded_file(self, filename, mimetype, printit, openit, filesize, options):
         filelog("do_process_downloaded_file%s", (filename, mimetype, printit, openit, filesize, options))
@@ -456,7 +455,7 @@ class FileTransferHandler(FileTransferAttributes):
             if DELETE_PRINTER_FILE:
                 try:
                     os.unlink(filename)
-                except (OSError, IOError):
+                except OSError:
                     printlog("failed to delete print job file '%s'", filename)
         if not printer:
             printlog.error("Error: the printer name is missing")
@@ -562,6 +561,10 @@ class FileTransferHandler(FileTransferAttributes):
         return True
 
 
+    def send_request_file(self, filename, openit=True):
+        self.send("request-file", filename, openit)
+
+
     def _process_open_url(self, packet):
         url, send_id = packet[1:3]
         if not self.open_url:
@@ -778,7 +781,7 @@ class FileTransferHandler(FileTransferAttributes):
         #convert str to utf8 bytes:
         try:
             base = basefilename.encode("utf8")
-        except:
+        except UnicodeEncodeError:
             base = strtobytes(basefilename)
         self.send("send-file", base, mimetype, printit, openit, filesize, cdata, options, send_id)
         return True
@@ -800,12 +803,12 @@ class FileTransferHandler(FileTransferAttributes):
         #send some more file data
         filelog("ack-file-chunk: %s", packet[1:])
         chunk_id, state, error_message, chunk = packet[1:5]
-        chunk_id = bytestostr(chunk_id)
         if not state:
             filelog.error("Error: remote end is cancelling the file transfer:")
             filelog.error(" %s", error_message)
             del self.send_chunks_in_progress[chunk_id]
             return
+        chunk_id = bytestostr(chunk_id)
         chunk_state = self.send_chunks_in_progress.get(chunk_id)
         if not chunk_state:
             filelog.error("Error: cannot find the file transfer id '%s'", nonl(chunk_id))
