@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -15,15 +15,17 @@ import atexit
 import traceback
 
 from xpra.scripts.main import info, warn, no_gtk, validate_encryption, parse_env, configure_env
-from xpra.scripts.config import InitException, FALSE_OPTIONS
+from xpra.scripts.config import InitException, FALSE_OPTIONS, parse_bool
+from xpra.common import CLOBBER_USE_DISPLAY, CLOBBER_UPGRADE
+from xpra.exit_codes import EXIT_VFB_ERROR, EXIT_OK
 from xpra.os_util import (
-    SIGNAMES, POSIX, WIN32, OSX, PYTHON3,
+    SIGNAMES, POSIX, WIN32, OSX,
     FDChangeCaptureContext,
     force_quit,
     get_username_for_uid, get_home_for_uid, get_shell_for_uid, getuid, setuidgid,
     get_hex_uuid, get_status_output, strtobytes, bytestostr, get_util_logger, osexpand,
     )
-from xpra.util import envbool, unsetenv
+from xpra.util import envbool, unsetenv, noerr
 from xpra.platform.dotxpra import DotXpra
 
 
@@ -106,7 +108,7 @@ def validate_pixel_depth(pixel_depth, starting_desktop=False):
     try:
         pixel_depth = int(pixel_depth)
     except ValueError:
-        raise InitException("invalid value '%s' for pixel depth, must be a number" % pixel_depth)
+        raise InitException("invalid value '%s' for pixel depth, must be a number" % pixel_depth) from None
     if pixel_depth==0:
         pixel_depth = 24
     if pixel_depth not in (8, 16, 24, 30):
@@ -124,9 +126,6 @@ def display_name_check(display_name):
     n = display_name[1:].split(".")[0]    #ie: ":0.0" -> "0"
     try:
         dno = int(n)
-    except (ValueError, TypeError):
-        raise InitException("invalid display number %r" % n)
-    else:
         if 0<=dno<10:
             warn("WARNING: low display number: %s" % dno)
             warn(" You are attempting to run the xpra server")
@@ -134,6 +133,8 @@ def display_name_check(display_name):
             warn(" This is generally not what you want.")
             warn(" You should probably use a higher display number")
             warn(" just to avoid any confusion and this warning message.")
+    except IOError:
+        pass
 
 
 def print_DE_warnings():
@@ -308,7 +309,7 @@ def set_server_features(opts):
     server_features.mmap            = b(opts.mmap)
     server_features.input_devices   = not opts.readonly and impcheck("keyboard")
     server_features.commands        = impcheck("server.control_command")
-    server_features.dbus            = opts.dbus_proxy and impcheck("dbus")
+    server_features.dbus            = opts.dbus_proxy and impcheck("dbus", "server.dbus")
     server_features.encoding        = impcheck("codecs")
     server_features.logging         = b(opts.remote_logging)
     #server_features.network_state   = ??
@@ -345,6 +346,22 @@ def run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=None
         gc.collect()
 
 
+def verify_display(xvfb=None, display_name=None, shadowing=False, log_errors=True):
+    #check that we can access the X11 display:
+    from xpra.x11.vfb_util import verify_display_ready
+    if not verify_display_ready(xvfb, display_name, shadowing, log_errors):
+        return 1
+    from xpra.log import Logger
+    log = Logger("screen", "x11")
+    log("X11 display is ready")
+    no_gtk()
+    from xpra.x11.gtk_x11.gdk_display_source import verify_gdk_display
+    display = verify_gdk_display(display_name)
+    if not display:
+        return 1
+    log("GDK can access the display")
+    return 0
+
 def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=None):
     assert mode in (
         "start", "start-desktop",
@@ -368,34 +385,24 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if k.startswith("DBUS_"):
             del os.environ[k]
 
+    use_display = parse_bool("use-display", opts.use_display)
     starting  = mode == "start"
     starting_desktop = mode == "start-desktop"
     upgrading = mode == "upgrade"
     upgrading_desktop = mode == "upgrade-desktop"
     shadowing = mode == "shadow"
     proxying  = mode == "proxy"
-    clobber   = int(upgrading or upgrading_desktop) | int(opts.use_display)*2
-    start_vfb = not (shadowing or proxying or clobber)
 
     if not proxying and POSIX and not OSX:
         #we don't support wayland servers,
         #so make sure GDK will use the X11 backend:
-        from xpra.os_util import saved_env
-        saved_env["GDK_BACKEND"] = "x11"
         os.environ["GDK_BACKEND"] = "x11"
 
-    has_child_arg = (
-            opts.start_child or
-            opts.start_child_on_connect or
-            opts.start_child_after_connect or
-            opts.start_child_on_last_client_exit
-            )
     if proxying or upgrading or upgrading_desktop:
         #when proxying or upgrading, don't exec any plain start commands:
-        opts.start = []
-        opts.start_child = []
+        opts.start = opts.start_child = []
     elif opts.exit_with_children:
-        assert has_child_arg, "exit-with-children was specified but start-child* is missing!"
+        assert opts.start_child, "exit-with-children was specified but start-child is missing!"
     elif opts.start_child:
         warn("Warning: the 'start-child' option is used,")
         warn(" but 'exit-with-children' is not enabled,")
@@ -403,14 +410,16 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
 
     if opts.bind_rfb and (proxying or starting):
         get_util_logger().warn("Warning: bind-rfb sockets cannot be used with '%s' mode" % mode)
-        opts.bind_rfb = ""
+        opts.bind_rfb = []
 
     if not shadowing and not starting_desktop:
         opts.rfb_upgrade = 0
 
     if upgrading or upgrading_desktop or shadowing:
         #there should already be one running
-        opts.pulseaudio = False
+        #so change None ('auto') to False
+        if opts.pulseaudio is None:
+            opts.pulseaudio = False
 
     #get the display name:
     if shadowing and not extra_args:
@@ -428,7 +437,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             error_cb("too many extra arguments (%i): only expected a display number" % len(extra_args))
         if len(extra_args) == 1:
             display_name = extra_args[0]
-            if not shadowing and not proxying and not opts.use_display:
+            if not shadowing and not proxying and not upgrading and not use_display:
                 display_name_check(display_name)
         else:
             if proxying:
@@ -445,7 +454,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
                         break
                 if not display_name:
                     error_cb("you must specify a free virtual display name to use with the proxy server")
-            elif opts.use_display:
+            elif use_display:
                 #only use automatic guess for xpra displays and not X11 displays:
                 display_name = guess_xpra_display(opts.socket_dir, opts.socket_dirs)
             else:
@@ -454,7 +463,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
                 display_name = 'S' + str(os.getpid())
 
     if not (shadowing or proxying or upgrading or upgrading_desktop) and \
-    opts.exit_with_children and not has_child_arg:
+    opts.exit_with_children and not opts.start_child:
         error_cb("--exit-with-children specified without any children to spawn; exiting immediately")
 
     atexit.register(run_cleanups)
@@ -476,9 +485,6 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     gid = int(opts.gid)
     username = get_username_for_uid(uid)
     home = get_home_for_uid(uid)
-    xauth_data = None
-    if start_vfb:
-        xauth_data = get_hex_uuid()
     ROOT = POSIX and getuid()==0
 
     protected_fds = []
@@ -504,6 +510,12 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             stderr.write(" %s\n" % e)
             del e
 
+    clobber = int(upgrading or upgrading_desktop)*CLOBBER_UPGRADE | int(use_display or 0)*CLOBBER_USE_DISPLAY
+    start_vfb = not (shadowing or proxying or clobber)
+    xauth_data = None
+    if start_vfb:
+        xauth_data = get_hex_uuid()
+
     # if pam is present, try to create a new session:
     pam = None
     PAM_OPEN = POSIX and envbool("XPRA_PAM_OPEN", ROOT and uid!=0)
@@ -548,11 +560,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
 
     #get XDG_RUNTIME_DIR from env options,
     #which may not be have updated os.environ yet when running as root with "--uid="
-    xrd = parse_env(opts.env).get("XDG_RUNTIME_DIR", "")
-    if OSX and not xrd:
-        xrd = osexpand("~/.xpra", uid=uid, gid=gid)
-        os.environ["XDG_RUNTIME_DIR"] = xrd
-    xrd = os.path.abspath(xrd)
+    xrd = os.path.abspath(parse_env(opts.env).get("XDG_RUNTIME_DIR", ""))
     if ROOT and (uid>0 or gid>0):
         #we're going to chown the directory if we create it,
         #ensure this cannot be abused, only use "safe" paths:
@@ -576,7 +584,8 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
 
     import datetime
     extra_expand = {"TIMESTAMP" : datetime.datetime.now().strftime("%Y%m%d-%H%M%S")}
-    if start_vfb or opts.daemon:
+    log_to_file = opts.daemon or os.environ.get("XPRA_LOG_TO_FILE", "")=="1"
+    if start_vfb or log_to_file:
         #we will probably need a log dir
         #either for the vfb, or for our own log file
         log_dir = opts.log_dir or ""
@@ -589,26 +598,33 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if "XPRA_LOG_DIR" not in os.environ:
             os.environ["XPRA_LOG_DIR"] = log_dir
 
-        if opts.daemon:
-            from xpra.server.server_util import select_log_file, open_log_file, redirect_std_to_log
-            log_filename0 = osexpand(select_log_file(log_dir, opts.log_file, display_name),
-                                     username, uid, gid, extra_expand)
-            logfd = open_log_file(log_filename0)
-            if ROOT and (uid>0 or gid>0):
-                try:
-                    os.fchown(logfd, uid, gid)
-                except (OSError, IOError):
-                    pass
-            stdout, stderr = redirect_std_to_log(logfd, *protected_fds)
+    if log_to_file:
+        from xpra.server.server_util import select_log_file, open_log_file, redirect_std_to_log
+        log_filename0 = osexpand(select_log_file(log_dir, opts.log_file, display_name),
+                                 username, uid, gid, extra_expand)
+        if os.path.exists(log_filename0) and not display_name.startswith("S"):
+            #don't overwrite the log file just yet,
+            #as we may still fail to start
+            log_filename0 += ".new"
+        logfd = open_log_file(log_filename0)
+        if POSIX and ROOT and (uid>0 or gid>0):
             try:
-                stderr.write("Entering daemon mode; "
-                         + "any further errors will be reported to:\n"
-                         + ("  %s\n" % log_filename0))
-                stderr.flush()
-            except IOError:
-                #we tried our best, logging another error won't help
-                pass
-            os.environ["XPRA_SERVER_LOG"] = log_filename0
+                os.fchown(logfd, uid, gid)
+            except OSError as e:
+                noerr(stderr.write, "failed to chown the log file '%s'\n" % log_filename0)
+                noerr(stderr.flush)
+        stdout, stderr = redirect_std_to_log(logfd, *protected_fds)
+        noerr(stderr.write, "Entering daemon mode; "
+                     + "any further errors will be reported to:\n"
+                     + ("  %s\n" % log_filename0))
+        noerr(stderr.flush)
+        os.environ["XPRA_SERVER_LOG"] = log_filename0
+    else:
+        #server log does not exist:
+        try:
+            del os.environ["XPRA_SERVER_LOG"]
+        except KeyError:
+            pass
 
     #warn early about this:
     if (starting or starting_desktop) and desktop_display and opts.notifications and not opts.dbus_launch:
@@ -649,6 +665,13 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     xvfb = None
     xvfb_pid = None
     uinput_uuid = None
+    if start_vfb and use_display is None:
+        #use-display='auto' so we have to figure out
+        #if we have to start the vfb or not:
+        if not display_name:
+            use_display = False
+        else:
+            start_vfb = verify_display(None, display_name, log_errors=False)!=0
     if start_vfb:
         assert not proxying and xauth_data
         pixel_depth = validate_pixel_depth(opts.pixel_depth, starting_desktop)
@@ -670,8 +693,8 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
         if display_name!=odisplay_name and pam:
             pam.set_items({"XDISPLAY" : display_name})
 
-        def check_xvfb():
-            return check_xvfb_process(xvfb)
+        def check_xvfb(timeout=0):
+            return check_xvfb_process(xvfb, timeout=timeout)
     else:
         if POSIX and clobber:
             #if we're meant to be using a private XAUTHORITY file,
@@ -681,7 +704,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             if os.path.exists(xauthority):
                 log("found XAUTHORITY=%s", xauthority)
                 os.environ["XAUTHORITY"] = xauthority
-        def check_xvfb():
+        def check_xvfb(timeout=0):
             return True
 
     if POSIX and not OSX and displayfd>0:
@@ -698,12 +721,13 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             log.error(" %s", str(e) or type(e))
             del e
 
+    if not check_xvfb(1):
+        noerr(stderr.write, "vfb failed to start, exiting\n")
+        return EXIT_VFB_ERROR
+
+    if WIN32 and os.environ.get("XPRA_LOG_FILENAME"):
+        os.environ["XPRA_SERVER_LOG"] = os.environ["XPRA_LOG_FILENAME"]
     if opts.daemon:
-        def noerr(fn, *args):
-            try:
-                fn(*args)
-            except Exception:
-                pass
         log_filename1 = osexpand(select_log_file(log_dir, opts.log_file, display_name),
                                  username, uid, gid, extra_expand)
         if log_filename0 != log_filename1:
@@ -721,19 +745,13 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             noerr(stderr.flush)
         noerr(stdout.close)
         noerr(stderr.close)
-    else:
-        #server log does not exist:
-        try:
-            del os.environ["XPRA_SERVER_LOG"]
-        except KeyError:
-            pass
     #we should not be using stdout or stderr from this point:
     del stdout
     del stderr
 
     if not check_xvfb():
-        #xvfb problem: exit now
-        return  1
+        noerr(stderr.write, "vfb failed to start, exiting\n")
+        return EXIT_VFB_ERROR
 
     #create devices for vfb if needed:
     devices = {}
@@ -797,22 +815,13 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
             if dbus_env:
                 os.environ.update(dbus_env)
 
-    display = None
     if not proxying:
         if POSIX and not OSX:
             no_gtk()
             if starting or starting_desktop or shadowing:
-                #check that we can access the X11 display:
-                from xpra.x11.vfb_util import verify_display_ready
-                if not verify_display_ready(xvfb, display_name, shadowing):
-                    return 1
-                log("X11 display is ready")
-                no_gtk()
-                from xpra.x11.gtk_x11.gdk_display_source import verify_gdk_display
-                display = verify_gdk_display(display_name)
-                if not display:
-                    return 1
-                log("GDK can access the display")
+                r = verify_display(xvfb, display_name, shadowing)
+                if r:
+                    return r
         #on win32, this ensures that we get the correct screen size to shadow:
         from xpra.platform.gui import init as gui_init
         log("gui_init()")
@@ -826,18 +835,18 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
                                         opts.mmap_group, opts.socket_permissions,
                                         username, uid, gid)
     netlog("setting up local sockets: %s", local_sockets)
-    sockets += local_sockets
+    sockets.update(local_sockets)
     if POSIX and (starting or upgrading or starting_desktop or upgrading_desktop):
         #all unix domain sockets:
         ud_paths = [sockpath for stype, _, sockpath, _ in local_sockets if stype=="unix-domain"]
         if ud_paths:
-            os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
             #choose one so our xdg-open override script can use to talk back to us:
             if opts.forward_xdg_open:
                 for x in ("/usr/libexec/xpra", "/usr/lib/xpra"):
                     xdg_override = os.path.join(x, "xdg-open")
                     if os.path.exists(xdg_override):
                         os.environ["PATH"] = x+os.pathsep+os.environ.get("PATH", "")
+                        os.environ["XPRA_SERVER_SOCKET"] = ud_paths[0]
                         break
         else:
             log.warn("Warning: no local server sockets,")
@@ -909,7 +918,7 @@ def do_run_server(error_cb, opts, mode, xpra_file, extra_args, desktop_display=N
     except KeyboardInterrupt:
         log.info("stopping on KeyboardInterrupt")
         app.cleanup()
-        return 0
+        return EXIT_OK
     except Exception:
         log.error("server error", exc_info=True)
         app.cleanup()
