@@ -9,7 +9,7 @@ import os.path
 
 from xpra.util import parse_scaling_value, from0to100
 from xpra.server.control_command import ArgsControlCommand, ControlError
-from xpra.os_util import load_binary_file, bytestostr, strtobytes
+from xpra.os_util import load_binary_file
 from xpra.util import csv
 from xpra.scripts.config import parse_bool, FALSE_OPTIONS, TRUE_OPTIONS
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
@@ -93,7 +93,7 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("scaling",               "set a specific scaling value",     min_args=1, validation=[parse_scaling_value]),
             ArgsControlCommand("auto-refresh",          "set a specific auto-refresh value", min_args=1, validation=[float]),
             ArgsControlCommand("refresh",               "refresh some or all windows",      min_args=0),
-            ArgsControlCommand("encoding",              "picture encoding",                 min_args=2),
+            ArgsControlCommand("encoding",              "picture encoding",                 min_args=1, max_args=1),
             ArgsControlCommand("video-region-enabled",  "enable video region",              min_args=2, max_args=2, validation=[int, parse_boolean_value]),
             ArgsControlCommand("video-region-detection","enable video detection",           min_args=2, max_args=2, validation=[int, parse_boolean_value]),
             ArgsControlCommand("video-region-exclusion-zones","set window regions to exclude from video regions: 'WID,(x,y,w,h),(x,y,w,h),..', ie: '1 (0,10,100,20),(200,300,20,20)'",  min_args=2, max_args=2, validation=[int, parse_4intlist]),
@@ -101,7 +101,7 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("reset-video-region",    "reset video region heuristics",    min_args=1, max_args=1, validation=[int]),
             ArgsControlCommand("lock-batch-delay",      "set a specific batch delay for a window",       min_args=2, max_args=2, validation=[int, int]),
             ArgsControlCommand("unlock-batch-delay",    "let the heuristics calculate the batch delay again for a window (following a 'lock-batch-delay')",  min_args=1, max_args=1, validation=[int]),
-            ArgsControlCommand("remove-window-filters", "remove all window filters",        min_args=0, max_args=1),
+            ArgsControlCommand("remove-window-filters", "remove all window filters",        min_args=0, max_args=0),
             ArgsControlCommand("add-window-filter",     "add a window filter",              min_args=4, max_args=5),
             ):
             cmd.do_run = getattr(self, "control_command_%s" % cmd.name.replace("-", "_"))
@@ -241,13 +241,6 @@ class ServerBaseControlCommands(StubServerMixin):
         return "url sent to %i clients" % clients
 
     def control_command_send_file(self, filename, openit="open", client_uuids="*", maxbitrate=0):
-        #we always get the values as strings from the command interface,
-        #but those may actually be utf8 encoded binary strings,
-        #so we may have to do an ugly roundtrip:
-        try:
-            filename = strtobytes(filename).decode("utf8")
-        except:
-            filename = bytestostr(filename)
         openit = str(openit).lower() in ("open", "true", "1")
         return self.do_control_file_command("send file", client_uuids, filename, "file_transfer", (False, openit))
 
@@ -266,6 +259,12 @@ class ServerBaseControlCommands(StubServerMixin):
         sources = self._control_get_sources(client_uuids)
         if not sources:
             raise ControlError("no clients found matching: %s" % client_uuids)
+        def checksize(file_size):
+            file_size_MB = file_size//1024//1024
+            if file_size_MB>self.file_transfer.file_size_limit:
+                raise ControlError("file '%s' is too large: %iMB (limit is %iMB)" % (
+                    filename, file_size_MB, self.file_transfer.file_size_limit))
+
         #find the file and load it:
         actual_filename = os.path.abspath(os.path.expanduser(filename))
         try:
@@ -273,15 +272,17 @@ class ServerBaseControlCommands(StubServerMixin):
             log("os.stat(%s)=%s", actual_filename, stat)
         except os.error:
             log("os.stat(%s)", actual_filename, exc_info=True)
+        else:
+            checksize(stat.st_size)
         if not os.path.exists(actual_filename):
             raise ControlError("file '%s' does not exist" % filename)
         data = load_binary_file(actual_filename)
+        if data is None:
+            raise ControlError("failed to load '%s'" % actual_filename)
         #verify size:
         file_size = len(data)
         file_size_MB = file_size//1024//1024
-        if file_size_MB>self.file_transfer.file_size_limit:
-            raise ControlError("file '%s' is too large: %iMB (limit is %iMB)" % (
-                filename, file_size_MB, self.file_transfer.file_size_limit))
+        checksize(file_size)
         #send it to each client:
         for ss in sources:
             #ie: ServerSource.file_transfer (found in FileTransferAttributes)
@@ -296,7 +297,7 @@ class ServerBaseControlCommands(StubServerMixin):
         return "%s of '%s' to %s initiated" % (command_type, filename, client_uuids)
 
 
-    def control_command_remove_window_filters(self, client_uuids=""):
+    def control_command_remove_window_filters(self):
         #modify the existing list object,
         #which is referenced by all the sources
         l = len(self.window_filters)
@@ -350,7 +351,7 @@ class ServerBaseControlCommands(StubServerMixin):
 
     def control_command_client(self, *args):
         self.all_send_client_command(*args)
-        return "client control command '%s' forwarded to clients" % str(args)
+        return "client control command %s forwarded to clients" % str(args)
 
     def control_command_client_property(self, wid, uuid, prop, value, conv=None):
         wid = int(wid)
@@ -370,7 +371,6 @@ class ServerBaseControlCommands(StubServerMixin):
         log.info("changed session name: %s", self.session_name)
         #self.all_send_client_command("name", name)    not supported by any clients, don't bother!
         self.setting_changed("session_name", name)
-        self.mdns_update()
         return "session name set to %s" % name
 
     def _control_windowsources_from_args(self, *args):
