@@ -1,17 +1,18 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.util import csv, nonl, envbool, repr_ellipsized
+from gi.repository import Gdk
+
+from xpra.util import csv, nonl, envbool
 from xpra.os_util import bytestostr
 from xpra.gtk_common.keymap import get_gtk_keymap
-from xpra.gtk_common.gtk_util import keymap_get_for_display, display_get_default, get_default_root_window
+from xpra.gtk_common.gtk_util import get_default_root_window
 from xpra.gtk_common.error import xsync, xlog
-from xpra.gtk_common.gobject_compat import import_gdk, is_gtk3
 from xpra.keyboard.mask import DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names
 from xpra.server.keyboard_config_base import KeyboardConfigBase
 from xpra.x11.gtk_x11.keys import grok_modifier_map
@@ -19,7 +20,7 @@ from xpra.x11.xkbhelper import (
     do_set_keymap, set_all_keycodes, set_keycode_translation,
     get_modifiers_from_meanings, get_modifiers_from_keycodes,
     clear_modifiers, set_modifiers, map_missing_modifiers,
-    clean_keyboard_state, get_keycode_mappings,
+    clean_keyboard_state,
     DEBUG_KEYSYMS,
     )
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
@@ -30,8 +31,6 @@ log = Logger("keyboard")
 X11Keyboard = X11KeyboardBindings()
 
 MAP_MISSING_MODIFIERS = envbool("XPRA_MAP_MISSING_MODIFIERS", True)
-SIMULATE_MODIFIERS = envbool("XPRA_SIMULATE_MODIFIERS", True)
-SHIFT_LOCK = envbool("XPRA_SHIFT_LOCK", False)
 
 ALL_X11_MODIFIERS = {
                     "shift"     : 0,
@@ -44,7 +43,6 @@ ALL_X11_MODIFIERS = {
                     "mod5"      : 7
                     }
 
-
 class KeyboardConfig(KeyboardConfigBase):
     def __init__(self):
         KeyboardConfigBase.__init__(self)
@@ -56,7 +54,7 @@ class KeyboardConfig(KeyboardConfigBase):
         self.xkbmap_mod_managed = []
         self.xkbmap_mod_pointermissing = []
         self.xkbmap_mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
-        self.xkbmap_keycodes = []
+        self.xkbmap_keycodes = ()
         self.xkbmap_x11_keycodes = []
         self.xkbmap_layout = None
         self.xkbmap_variant = None
@@ -77,7 +75,7 @@ class KeyboardConfig(KeyboardConfigBase):
     def __repr__(self):
         return "KeyboardConfig(%s / %s / %s)" % (self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options)
 
-    def get_info(self):
+    def get_info(self) -> dict:
         info = KeyboardConfigBase.get_info(self)
         #keycodes:
         if self.keycode_translation:
@@ -137,7 +135,7 @@ class KeyboardConfig(KeyboardConfigBase):
     def parse_options(self, props):
         """ used by both process_hello and process_keymap
             to set the keyboard attributes """
-        KeyboardConfigBase.parse_options(self, props)
+        super().parse_options(props)
         modded = {}
         def parse_option(name, parse_fn, *parse_args):
             prop = "xkbmap_%s" % name
@@ -150,13 +148,13 @@ class KeyboardConfig(KeyboardConfigBase):
         for x in ("print", "query"):
             parse_option(x, props.strget)
         #lists:
-        parse_option("keycodes", props.listget)
+        parse_option("keycodes", props.tupleget)
         #dicts:
         for x in ("mod_meanings", "x11_keycodes", "query_struct"):
             parse_option(x, props.dictget, {})
         #lists of strings:
         for x in ("mod_managed", "mod_pointermissing"):
-            parse_option(x, props.strlistget)
+            parse_option(x, props.strtupleget)
         parse_option("raw", props.boolget)
         #older clients don't specify if they support layout groups safely
         #(MS Windows clients used base-1)
@@ -223,26 +221,22 @@ class KeyboardConfig(KeyboardConfigBase):
     def compute_modifier_keynames(self):
         self.keycodes_for_modifier_keynames = {}
         self.xkbmap_mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
-        display = display_get_default()
-        keymap = keymap_get_for_display(display)
-        gdk = import_gdk()
+        display = Gdk.Display.get_default()
+        keymap = Gdk.Keymap.get_for_display(display)
         if self.keynames_for_mod:
             for modifier, keynames in self.keynames_for_mod.items():
                 for keyname in keynames:
                     if keyname in DEFAULT_MODIFIER_NUISANCE_KEYNAMES:
                         self.xkbmap_mod_nuisance.add(modifier)
-                    keyval = gdk.keyval_from_name(bytestostr(keyname))
+                    keyval = Gdk.keyval_from_name(bytestostr(keyname))
                     if keyval==0:
                         log.error("Error: no keyval found for keyname '%s' (modifier '%s')", keyname, modifier)
                         return  []
                     entries = keymap.get_entries_for_keyval(keyval)
                     if entries:
                         keycodes = []
-                        if is_gtk3():
-                            if entries[0] is True:
-                                keycodes = [entry.keycode for entry in entries[1]]
-                        else:
-                            keycodes = [entry[0] for entry in entries]
+                        if entries[0] is True:
+                            keycodes = [entry.keycode for entry in entries[1]]
                         for keycode in keycodes:
                             l = self.keycodes_for_modifier_keynames.setdefault(keyname, [])
                             if keycode not in l:
@@ -289,7 +283,7 @@ class KeyboardConfig(KeyboardConfigBase):
             log.error("Error: compute_client_modifier_keycodes: %s" % e, exc_info=True)
 
     def compute_modifier_map(self):
-        self.modifier_map = grok_modifier_map(display_get_default(), self.xkbmap_mod_meanings)
+        self.modifier_map = grok_modifier_map(Gdk.Display.get_default(), self.xkbmap_mod_meanings)
         log("modifier_map(%s)=%s", self.xkbmap_mod_meanings, self.modifier_map)
 
 
@@ -374,26 +368,17 @@ class KeyboardConfig(KeyboardConfigBase):
             clean_keyboard_state()
             self.update_keycode_mappings()
 
+
     def add_gtk_keynames(self):
         #add the keynames we find via gtk
         #since we may rely on finding those keynames from the client
         #(used with non native keymaps)
+        log("add_gtk_keynames() gtk keymap=%s", get_gtk_keymap())
         for _, keyname, keycode, _, _ in get_gtk_keymap():
             if keyname not in self.keycode_translation:
                 self.keycode_translation[keyname] = keycode
                 if keyname in DEBUG_KEYSYMS:
                     log.info("add_gtk_keynames: %s=%s", keyname, keycode)
-        #some keys may still be missing with GTK2:
-        for keyname, equiv in {
-            "dead_perispomeni"  : "dead_tilde",
-            }.items():
-            if keyname not in self.keycode_translation:
-                keycode = self.keycode_translation.get(equiv)
-                if keycode:
-                    self.keycode_translation[keyname] = keycode
-                    if keyname in DEBUG_KEYSYMS:
-                        log.info("add_gtk_keynames: %s=%s", keyname, keycode)
-
 
     def set_default_keymap(self):
         """ assign a default keymap based on the current X11 server keymap
@@ -405,7 +390,7 @@ class KeyboardConfig(KeyboardConfigBase):
         with xsync:
             clean_keyboard_state()
             #keycodes:
-            keycode_to_keynames = get_keycode_mappings()
+            keycode_to_keynames = X11Keyboard.get_keycode_mappings()
             self.keycode_translation = {}
             #prefer keycodes that don't use the lowest level+mode:
             default_for_keyname = {}
@@ -443,10 +428,10 @@ class KeyboardConfig(KeyboardConfigBase):
             self.update_keycode_mappings()
 
     def update_keycode_mappings(self):
-        self.keycode_mappings = get_keycode_mappings()
+        self.keycode_mappings = X11Keyboard.get_keycode_mappings()
 
 
-    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, keystr, group):
+    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, group):
         if not self.enabled:
             log("ignoring keycode since keyboard is turned off")
             return -1, group
@@ -454,46 +439,6 @@ class KeyboardConfig(KeyboardConfigBase):
             return -1, group
         if self.xkbmap_raw:
             return client_keycode, group
-        keycode = None
-        if self.xkbmap_query:
-            keycode = self.keycode_translation.get((client_keycode, keyname)) or client_keycode
-            log("get_keycode(%s, %s, %s)=%s (native keymap)", client_keycode, keyname, modifiers, keycode)
-        else:
-            if SIMULATE_MODIFIERS:
-                return self.do_get_keycode_new(client_keycode, keyname, pressed, modifiers, group)
-            #non-native: try harder to find matching keysym
-            #first, try to honour shift state:
-            lock = ("lock" in modifiers) and (SHIFT_LOCK or (bool(keystr) and keystr.isalpha()))
-            shift = ("shift" in modifiers) ^ lock
-            mode = 0
-            for mod in modifiers:
-                names = self.keynames_for_mod.get(mod, [])
-                for name in names:
-                    if name in ("ISO_Level3_Shift", "Mode_switch"):
-                        mode = 1
-                        break
-            level = int(shift) + int(mode)*2
-            levels = []
-            if group:
-                levels.append(level+4)
-            levels.append(level)
-            for i in range(4):
-                level = int(shift) + i*2
-                if level not in levels:
-                    levels.append(level)
-            for level in levels:
-                keycode = self.keycode_translation.get((keyname, level))
-                if keycode:
-                    log("get_keycode(%s, %s, %s, %s)=%i (level=%i, shift=%s, mode=%i)",
-                        client_keycode, keyname, modifiers, group, keycode, level, shift, mode)
-                    break
-            if keycode is None:
-                keycode = self.keycode_translation.get(keyname, -1)
-                log("get_keycode(%s, %s)=%i (keyname translation)", client_keycode, keyname, keycode)
-        return keycode, group
-
-    def do_get_keycode_new(self, client_keycode, keyname, pressed, modifiers, group):
-        #non-native: try harder to find matching keysym
         def kmlog(msg, *args):
             if keyname in DEBUG_KEYSYMS:
                 l = log.info
@@ -502,85 +447,98 @@ class KeyboardConfig(KeyboardConfigBase):
             l(msg, *args)
         def klog(msg, *args):
             kmlog("do_get_keycode%s"+msg, (client_keycode, keyname, pressed, modifiers, group), *args)
-        #first, try to honour shift state:
-        lock = ("lock" in modifiers) and (SHIFT_LOCK or (bool(keystr) and keystr.isalpha()))
-        shift = ("shift" in modifiers) ^ lock
-        mode = 0
-        numlock = 0
-        numlock_modifier = None
-        for mod, keynames in self.keynames_for_mod.items():
-            if "Num_Lock" in keynames:
-                numlock_modifier = mod
-                break
-        for mod in modifiers:
-            names = self.keynames_for_mod.get(mod, [])
-            if "Num_Lock" in names:
-                numlock = 1
-            for name in names:
-                if name in ("ISO_Level3_Shift", "Mode_switch"):
-                    mode = 1
-                    break
-        level = int(shift) + int(mode)*2
-        levels = []
-        #try to preserve the mode (harder to toggle):
-        for m in (int(bool(mode)), int(not mode)):
-            #try to preserve shift state:
-            for s in (int(bool(shift)), int(not shift)):
-                #group is comparatively easier to toggle (one function call):
-                for g in (int(bool(group)), int(not group)):
-                    level = int(g)*4 + int(m)*2 + int(s)*1
-                    levels.append(level)
-        kmlog("will try levels: %s", levels)
+        keycode = None
         rgroup = group
-        for level in levels:
-            keycode = self.keycode_translation.get((keyname, level))
-            if keycode:
-                klog("=%i (level=%i, shift=%s, mode=%i)", keycode, level, shift, mode)
-                if self.xkbmap_raw:
+        if self.xkbmap_query:
+            keycode = self.keycode_translation.get((client_keycode, keyname)) or client_keycode
+            klog("=%s (native keymap)", keycode)
+        else:
+            """
+            from man xmodmap:
+            The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
+            hex or octal and can be determined by running the xev program).
+            Up to eight keysyms may be attached to a key, however the last four are not used in any major
+            X server implementation.
+            The first keysym is used when no modifier key is pressed in conjunction with this key,
+            the second with Shift, the third when the Mode_switch key is used with this key and
+            the fourth when both the Mode_switch and Shift keys are used.
+            """
+            #non-native: try harder to find matching keysym
+            #first, try to honour shift state:
+            shift = ("shift" in modifiers) ^ ("lock" in modifiers)
+            mode = 0
+            numlock = 0
+            numlock_modifier = None
+            for mod, keynames in self.keynames_for_mod.items():
+                if "Num_Lock" in keynames:
+                    numlock_modifier = mod
                     break
-                keysyms = self.keycode_mappings.get(keycode)
-                level0 = levels[0]
-                uq_keysyms = set(keysyms)
-                if len(uq_keysyms)<=1 or (len(keysyms)>level0 and keysyms[level0]==""):
-                    #if the keysym we would match for this keycode is 'NoSymbol',
-                    #then we can probably ignore it ('NoSymbol' shows up as "")
-                    kmlog("not toggling any modifiers state for keysyms=%s", keysyms)
+            for mod in modifiers:
+                names = self.keynames_for_mod.get(mod, [])
+                if "Num_Lock" in names:
+                    numlock = 1
+                for name in names:
+                    if name in ("ISO_Level3_Shift", "Mode_switch"):
+                        mode = 1
+                        break
+            levels = []
+            #try to preserve the mode (harder to toggle):
+            for m in (int(bool(mode)), int(not mode)):
+                #try to preserve shift state:
+                for s in (int(bool(shift)), int(not shift)):
+                    #group is comparatively easier to toggle (one function call):
+                    for g in (int(bool(group)), int(not group)):
+                        level = int(g)*4 + int(m)*2 + int(s)*1
+                        levels.append(level)
+            kmlog("will try levels: %s", levels)
+            for level in levels:
+                keycode = self.keycode_translation.get((keyname, level))
+                if keycode:
+                    keysyms = self.keycode_mappings.get(keycode)
+                    klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
+                    if self.xkbmap_raw:
+                        break
+                    level0 = levels[0]
+                    if len(keysyms)>level0 and keysyms[level0]=="":
+                        #if the keysym we would match for this keycode is 'NoSymbol',
+                        #then we can probably ignore it ('NoSymbol' shows up as "")
+                        kmlog("not toggling any modifiers state for keysyms=%s", keysyms)
+                        break
+                    def toggle_modifier(mod):
+                        keynames = self.keynames_for_mod.get(mod)
+                        if keyname in keynames:
+                            kmlog("not toggling '%s' since '%s' should deal with it", mod, keyname)
+                            #the keycode we're returning is for this modifier,
+                            #assume that this will end up doing what is needed
+                            return
+                        if mod in modifiers:
+                            kmlog("removing '%s' from modifiers", mod)
+                            modifiers.remove(mod)
+                        else:
+                            kmlog("adding '%s' to modifiers", mod)
+                            modifiers.append(mod)
+                    #keypad overrules shift state (see #2702):
+                    if keyname.startswith("KP_"):
+                        if numlock_modifier and not numlock:
+                            toggle_modifier(numlock_modifier)
+                    elif (level & 1) ^ shift:
+                        #shift state does not match
+                        toggle_modifier("shift")
+                    if int(bool(level & 2)) ^ mode:
+                        #try to set / unset mode:
+                        for mod, keynames in self.keynames_for_mod.items():
+                            if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
+                                #found mode switch modified
+                                toggle_modifier(mod)
+                                break
+                    rgroup = level//4
+                    if rgroup!=group:
+                        kmlog("switching group from %i to %i", group, rgroup)
                     break
-                def toggle_modifier(mod):
-                    keynames = self.keynames_for_mod.get(mod)
-                    if keyname in keynames:
-                        kmlog("not toggling '%s' since '%s' should deal with it", mod, keyname)
-                        #the keycode we're returning is for this modifier,
-                        #assume that this will end up doing what is needed
-                        return
-                    if mod in modifiers:
-                        kmlog("removing '%s' from modifiers", mod)
-                        modifiers.remove(mod)
-                    else:
-                        kmlog("adding '%s' to modifiers", mod)
-                        modifiers.append(mod)
-                #keypad overrules shift state (see #2702):
-                if keyname.startswith("KP_"):
-                    if numlock_modifier and not numlock:
-                        toggle_modifier(numlock_modifier)
-                elif (level & 1) ^ shift:
-                    #shift state does not match
-                    toggle_modifier("shift")
-                if int(bool(level & 2)) ^ mode:
-                    #try to set / unset mode:
-                    for mod, keynames in self.keynames_for_mod.items():
-                        if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
-                            #found mode switch modified
-                            toggle_modifier(mod)
-                            break
-                rgroup = level//4
-                if rgroup!=group:
-                    kmlog("switching group from %i to %i", group, rgroup)
-                break
-        #this should not find anything new?:
-        if keycode is None:
-            keycode = self.keycode_translation.get(keyname, -1)
-            klog("=%i (keyname translation)", keycode)
+            #this should not find anything new?:
+            if keycode is None:
+                keycode = self.keycode_translation.get(keyname, -1)
+                klog("=%i, %i (keyname translation)", keycode, rgroup)
         return keycode, rgroup
 
 
